#!/usr/bin/env python3

"""
Creates summary page from RapidPE/RIFT results
"""

__author__ = "Vinaya Valsan"

import os

import subprocess as sp
import numpy as np

from tabulate import tabulate
from glob import glob
from argparse import ArgumentParser
from urllib.parse import urlparse

import rapidpe_rift_pipe.postscript_utils as postutils

from rapidpe_rift_pipe.config import Config
from rapidpe_rift_pipe.profiling import write_css_file
from rapidpe_rift_pipe.utils import print_output

optp = ArgumentParser()
optp.add_argument("input_dir", help="path to event run dir")
optp.add_argument("--web-dir", default=None, help="path to web dir")
optp.add_argument("--output-dir", default=None, help="directory to save plots")
opts = optp.parse_args()
print("-----------Creating summary page--------------")
input_dir = opts.input_dir

if opts.web_dir:
    output_dir = opts.web_dir
else:
    output_dir = os.path.join(
        os.getenv("HOME"),
        f'public_html/RapidPE/{input_dir[input_dir.rfind("output/") + 7 :]}',
    )
os.makedirs(output_dir, exist_ok=True)
write_css_file(output_dir, "stylesheet.css")

if opts.output_dir:
    run_dir = opts.output_dir
else:
    run_dir = input_dir
summary_plot_dir = os.path.join(run_dir, "summary")

os.system(f"cp {summary_plot_dir}/* {output_dir}/")
print(f"Summary page will be saved in {output_dir}")
index_html_file = os.path.join(output_dir, "summarypage.html")

html_file = open(index_html_file, "w")


print_output(
    html_file,
    """
<html>
<head>
<link rel="stylesheet" href="stylesheet.css">
<script src="https://www.kryogenix.org/code/browser/sorttable/sorttable.js"></script>
</head>
""",
)

print_output(html_file, "<body>")
print_output(html_file, f"<h2>rundir = {run_dir}</h2>")

event_info = postutils.event_info(input_dir)

event_info_dict = event_info.load_event_info()

print_output(html_file, "<h1> Event Info  </h1>")
print_output(
    html_file,
    f"""
snr = {event_info_dict["snr"]} <br>
approximant = {event_info_dict["approximant"]} <br>
pipeline_recovered_param = {event_info_dict["intrinsic_param"]} <br>
event_time = {event_info_dict["event_time"]} <br>
""",
)
config = Config.load(os.path.join(input_dir, "Config.ini"))
is_event = config.event.mode in {"sid", "gid"}
if is_event:
    gracedb_url = urlparse(config.gracedb_url)
    if config.event.mode == "sid":
        event_id = config.event.superevent_id
        event_path = f"/superevents/{event_id}/view/"
    else:
        event_id = config.event.gracedb_id
        event_path = f"/events/{event_id}/view/"
    event_url = gracedb_url._replace(path=event_path).geturl()
    print_output(
        html_file, f'GraceDB url : <a href="{event_url}">{event_id}</a> <br>'
    )

filelist = glob(output_dir + "/grid*png")
print_output(html_file, "<h1> Grid Plots </h1>")
distance_coordinate = config.distance_coordinates
all_grid_fname = os.path.join(
    output_dir, f"grid_{distance_coordinate}_all.png"
)
print_output(html_file, f'<img src="{os.path.basename(all_grid_fname)}">')
distance_coordinate_filelist = glob(
    f"{output_dir}/grid_{distance_coordinate}_iteration-*.png"
)
for fname_full in sorted(distance_coordinate_filelist):
    fname = os.path.basename(fname_full)
    print_output(html_file, f'<img src="{fname}">')
remaining_filelist = (
    set(filelist) - set([all_grid_fname]) - set(distance_coordinate_filelist)
)
for fname_full in sorted(remaining_filelist):
    fname = os.path.basename(fname_full)
    print_output(html_file, f'<img src="{fname}">')

print_output(html_file, "<h1> Posterior Plots </h1>")


filelist = glob(output_dir + "/posterior*.png")
for fname_full in sorted(filelist):
    fname = os.path.basename(fname_full)
    print_output(html_file, f'<img src="{fname}">')

filelist = glob(output_dir + "/p_astro*png")
if filelist != []:
    print_output(html_file, "<h1> Pastro </h1>")
    for fname_full in sorted(filelist):
        fname = os.path.basename(fname_full)
        print_output(html_file, f"<br>{fname}")
        print_output(html_file, f'<img src="{fname}">')

filelist = glob(output_dir + "/skymap*png")
if len(filelist) != 0:
    print_output(html_file, "<h1> Skymaps </h1>")
    for fname_full in sorted(filelist):
        fname = os.path.basename(fname_full)
        print_output(html_file, f"<br>{fname}")
        print_output(html_file, f'<img src="{fname}">')


if config.cProfile:
    print_output(html_file, """<h1> Timing </h1> """)

    if os.path.exists(f"{summary_plot_dir}/cprofile.html"):
        filelist = np.sort(glob(output_dir + "/cprofile*hist*png"))
        for fname_full in sorted(filelist):
            fname = os.path.basename(fname_full)
            print_output(html_file, f'<img src="{fname}">')

    # Total job time:
    condor_submit_time = int(event_info_dict["condor_submit_time"])
    job_timing_file = os.path.join(input_dir, "job_timing.txt")
    iteration_completion_time = []
    with open(job_timing_file) as f:
        lines = f.readlines()
        for line_id, line in enumerate(lines):
            line_split = line.split(" ")
            level_complete_time = float(line_split[1])
            iteration_completion_time.append(level_complete_time)
            if line_id == 0:
                print_output(
                    html_file,
                    f'<br> <font size="+2"> iteration level {line_split[0]} took '
                    f"{level_complete_time-condor_submit_time} s </font>",
                )
            else:
                print_output(
                    html_file,
                    f'<br> <font size="+2"> iteration level {line_split[0]} took '
                    f"{level_complete_time-iteration_completion_time[line_id-1]} s </font>",
                )

    if os.path.exists(f"{summary_plot_dir}/cprofile.html"):
        print_output(
            html_file,
            "<br><a href='cprofile.html'>Detailed cProfile info for a single ILE job</a>",
        )
    print_output(html_file, "<h1> Config.ini </h1>")

with open(os.path.join(input_dir, "Config.ini")) as config_f:
    for line in config_f:
        if line[0] != "#" and len(line.strip()) > 0:
            if line[0] == "[":
                print_output(html_file, f"<br> <b> {line} </b>")
            else:
                print_output(html_file, f"<br> {line}")


print_output(html_file, "</body></html>")

html_file.close()



__author__ = "Sinead Walsh"

import ast,sys,os
import numpy as np

def convert_section_args_to_dict(cfg,section):
    #in principle this can be done with a single line, but i want to include sanity checks. 
#    string_of_dict = json.dumps(dict(cfg.items(section)))
    if not cfg.has_section(section):
        return {}

    dictin = dict(cfg.items(section))

    for opt,arg in list(dictin.items()):
#        print opt,arg
        if len(arg) > 1 and arg[0] is "[" and arg[-1] is "]":
            dictin[opt] = correct_list_string_formatting(arg)
        else:
            dictin[opt] = arg

    return dictin


def convert_dict_string_to_dict(input_dict_string):
    #in principle this can be done with a single line, but i want to include sanity checks. 
#    string_of_dict = json.dumps(dict(cfg.items(section)))

    dictin = ast.literal_eval(input_dict_string)

    for opt,arg in dictin.items():
        dictin[opt] = correct_list_string_formatting_if_list_string(arg)

    return dictin


#convert cfg options directly to command line arguments              
def convert_cfg_section_to_cmd_line(cfg,cfg_section):
    cmds_dict = dict(cfg.items(cfg_section))
    cmd_line = ""
    for name,arg in list(cmds_dict.items()):
        #check if it's meant to be a list and if yes convert it to format for ArgumentParser append option
        if len(arg) > 1 and arg[0] is "[" and arg[-1] is "]":
            arglist = ast.literal_eval(correct_list_string_formatting(arg))
            for subarg in arglist:
                cmd_line += " --"+name+"="+subarg
        else:
            cmd_line += " --"+name
            if len(arg) > 0:
                cmd_line +="="+arg
    return cmd_line

def convert_dict_to_cmd_line(cmds_dict):
    cmd_line = ""
    for name, arg in cmds_dict.items():
        #check if it's meant to be a list and if yes convert it to format for ArgumentParser append option
        if len(arg) > 1 and arg[0] is "[" and arg[-1] is "]":
            arglist = ast.literal_eval(correct_list_string_formatting(arg))
            for subarg in arglist:
                cmd_line += " --"+name+" "+subarg
        else:
            cmd_line += " --"+name
            if len(arg) > 0:
                cmd_line +=" "+arg
    return cmd_line


def convert_list_string_to_dict(list_string):
    output_dict = {}
    list_string = correct_list_string_formatting(list_string)
#    print "ls",list_string
    for subarg in ast.literal_eval(list_string):
#        print subarg
        tmp=subarg.split("=")
        output_dict[tmp[0]] = tmp[1]
    return output_dict

#make sure the format is ['arg1=opt1','arg2=opt2'], otherwise ast.literal eval wont know how to split it
#If the arg is a list, make sure the items wihtin the list are strings, otherwise each character will be split later 
def correct_list_string_formatting(list_string):
    output_str= ""
    list_string = list_string.replace("\n","")
    if not (len(list_string) > 1 and list_string[0] is "[" and list_string[-1] is "]"):
        #Check if theres just one arg=opt
        if list_string.count("=") == 1 and list_string.count("[") == 0  and list_string.count("]") == 0:
            if not "'" in list_string and not '"' in list_string:
                output_str = "['"+list_string+"']"
            else:
                output_str = "["+list_string+"]"
        else:
            print((len(list_string),list_string[0], list_string[-1:],list_string.split(),list_string.replace(" ","")[-1:]))
            sys.exit("ERROR: incorrect format. The variables should be in the form '[arg1=opt1,arg2=opt2]', not "+list_string)
    else:
        #First, make sure the format is ['arg1=opt1','arg2=opt2'], otherwise ast.literal eval wont know how to split it
        arg2 = list_string
        if not '"' in arg2 and not "'" in arg2:
            arg2 = arg2.replace('[','["')
            arg2 = arg2.replace(',','","')
            arg2 = arg2.replace(']','"]')
        output_str = arg2
    return output_str


def correct_list_string_formatting_if_list_string(list_string_or_other):
    if not isinstance(list_string_or_other, str):
        return list_string_or_other

    if list_string_or_other[0] != "[":
        return list_string_or_other

    if list_string_or_other[-1] != "]":
        return list_string_or_other

    return correct_list_string_formatting(list_string_or_other)


def get_html_str(plot_name_no_path):
    return '<img src="'+plot_name_no_path+'" style="float: left; width: 45%; margin-right: 1%; margin-bottom: 0.5em;">'
#    return '<img src="'+plot_name_no_path+'" style="float: left; width: 100%; margin-right: 1%; margin-bottom: 0.5em;">'

def html_newline():
    return '<p style="clear: both;">\n'

def check_switch_m1m2s1s2(intr_prm):
    if intr_prm["mass1"] < intr_prm["mass2"]:
        copy = intr_prm.copy()
        copy["mass2"] = intr_prm["mass1"]
        copy["mass1"] = intr_prm["mass2"]
        if "spin1z" in intr_prm:
            copy["spin2z"] = intr_prm["spin1z"]
            copy["spin1z"] = intr_prm["spin2z"]
        return copy
    return intr_prm
    

def get_mchirp_eta(m1,m2):
    mc = (((m1*m2)**(3./5.))/((m1+m2)**(1./5.)))
    eta = ((m1*m2)/((m1+m2)**2.))
    return mc,eta

def get_m1m2_from_mceta(Mc,eta):
    m1 = 0.5*Mc*eta**(-3./5.)*(1. + norm_sym_ratio(eta))
    m2 = 0.5*Mc*eta**(-3./5.)*(1. - norm_sym_ratio(eta))
    return m1,m2

def transform_s1zs2z_chi(m1, m2, s1z, s2z):
    return (m1 * s1z + m2 * s2z) / (m1 + m2)

def transform_s1zs2z_chi_eff_chi_a(mass1, mass2, spin1z, spin2z):
    #Copied from pycbc https://github.com/gwastro/pycbc/blob/master/pycbc/conversions.py         
    """ Returns the aligned mass-weighted spin difference from mass1, mass2,                     
    spin1z, and spin2z.                                                                          
    """
    chi_eff = (spin1z * mass1 + spin2z * mass2) / (mass1 + mass2)
    chi_a = (spin2z * mass2 - spin1z * mass1) / (mass2 + mass1)
    return chi_eff,chi_a

def transform_chi_eff_chi_a_s1zs2z(mass1, mass2, chi_eff, chi_a):
    """Returns spin1z.spin2z                                                                           
    """
    spin1z = (mass1 + mass2) / (2.0 * mass1) * (chi_eff - chi_a)
    spin2z = (mass1 + mass2) / (2.0 * mass2) * (chi_eff + chi_a)
    return spin1z,spin2z

def norm_sym_ratio(eta):
    assert np.all(eta <= 0.25)

    return np.sqrt(1 - 4. * eta)

def convert_injections_txt_to_objects(inj_file):
    output = []
    #inj = np.genfromtxt(inj_file,unpack=True,names=True)
    inj = np.genfromtxt(inj_file,names=True)
    #load the injections as a list of objects so they can easily be accessed in the same way
    # as SimInspiralTables.
    for ei,e in enumerate(inj["mass1"]):
        output.append(Injection(inj,ei))

    return output

def get_priors(param,val,prior_limits={}):

    islist = 1 if (isinstance(val,list) or isinstance(val,np.ndarray)) else 0

    weight = 1.0
    if param == "distance":
        min_dist_mpc = prior_limits["distance"][0] if "distance" in prior_limits else 0
        max_dist_mpc = prior_limits["distance"][1] if "distance" in prior_limits else 2000.0
        print(("Prior limits on distance",min_dist_mpc,max_dist_mpc))
        return val**2/(max_dist_mpc**3/3. - min_dist_mpc**3/3.)
    elif param == "inclination":
        return np.sin(val)/2.0
    elif param == "latitude":
        return np.cos(val)/2.0
    elif param == "phase" or param == "longitude" or param == "psi0":
        weight = 1.0/(2.0*np.pi)
    elif param == "polarization" or param == "phi0":
        weight = 1.0/(np.pi)
    elif param == "mchirp":
        minmc = 0.0
        maxmc=200.0
        if "mchirp" in prior_limits:
            minmc = prior_limits["mchirp"][0]
            maxmc = prior_limits["mchirp"][1]
        print(("Prior limits on mchirp",minmc,maxmc))
        weight = np.asarray([uniform_samp_prior(minmc,maxmc,e) for e in val]) if islist else uniform_samp_prior(minmc,maxmc,val)
        islist = 0
    elif param == "eta":
        weight = 4.0
    else:
        weight = 1.0

    return np.asarray([weight for e in val]) if islist else weight

def uniform_samp_prior(a, b, x):
    if x > a and x < b:
        return 1.0/(b-a)
    else:
        return 0

class Injection:
    def __init__(self,e,ei):
        self.mass1 = e["mass1"][ei]
        self.mass2 = e["mass2"][ei]
        self.spin1z = e["spin1z"][ei]
        self.spin2z = e["spin2z"][ei]
        self.latitude = e["latitude"][ei]
        self.longitude = e["longitude"][ei]
        self.distance = e["distance"][ei]
        self.inclination = e["inclination"][ei]
        self.psi0 = e["psi0"][ei]
        self.phi0 = e["phi0"][ei]
        self.snr = e["snr"][ei]
        if "geocent_end_time" in e.dtype.names:
            #WARNING: make sure the event_time saved in the txt file isn't missing the nanoseconds part
            self.geocent_end_time = e["geocent_end_time"][ei]
            self.geocent_end_time_ns = -1
        else:
            self.geocent_end_time = e["geocent_end_time_part1"][ei]
            self.geocent_end_time_ns = e["geocent_end_time_ns_part2"][ei]
        self.alpha6 = int(e["original_index"][ei])


def run_cmd(cmd,verbose=1):
    if verbose:
        print (cmd)
    exit_status = os.system(cmd)
    if exit_status != 0:
        print (cmd)
        sys.exit("ERROR: non zero exit status"+str(exit_status))


def construct_event_time_string(t_s,t_ns):
    if t_ns < 0:
        #if ns has been set to a value less than 0, it is because we shouldn't use it
        return str(t_s)
    t_s_s = str(t_s)
    t_ns_s = str(t_ns)
    while len(t_ns_s) < 9:
        t_ns_s += "0"
    t_str = t_s_s+"."+t_ns_s
    return t_str





#!/usr/bin/env python3

__author__ = "Caitlin Rose, Vinaya Valsan"

import json
import re
import glob
import h5py
import logging

import numpy as np
import matplotlib.pyplot as plt

from ligo.lw import utils, lsctables, ligolw
from scipy.stats import multinomial

from rapid_pe import amrlib
from rapid_pe.amrlib import VALID_TRANSFORMS_MASS
from rapid_pe.amrlib import INVERSE_TRANSFORMS_MASS
from rapid_pe.amrlib import BOUND_CHECK_MASS

import rapidpe_rift_pipe.jacobians as jacobians

logging.basicConfig(level=logging.INFO)


class event_info:
    def __init__(self, rundir):
        self.rundir = rundir

    def load_event_info(self):
        """
        read event_info_dict.txt
        """
        with open(self.rundir + "/event_info_dict.txt") as f:
            event_info_dict = json.load(f)
        return event_info_dict

    def load_injection_info(self):
        """
        read injection_info.txt
        """
        try:
            with open(self.rundir + "/injection_info.txt") as f:
                injection_info_dict = json.load(f)
            (
                injection_info_dict["chi_eff"],
                injection_info_dict["chi_a"],
            ) = amrlib.transform_s1zs2z_chi_eff_chi_a(
                injection_info_dict["mass1"],
                injection_info_dict["mass2"],
                injection_info_dict["spin1z"],
                injection_info_dict["spin2z"],
            )
        except FileNotFoundError:
            injection_info_dict = None
        return injection_info_dict

    def get_event_params(self):
        event_info_dict = self.load_event_info()
        intrinsic_param_event = event_info_dict["intrinsic_param"]
        mass1_event = float(
            re.search('mass1=(.+?)"', intrinsic_param_event).group(1)
        )
        mass2_event = float(
            re.search('mass2=(.+?)"', intrinsic_param_event).group(1)
        )
        event_params = {}
        event_params["mass1"] = mass1_event
        event_params["mass2"] = mass2_event
        try:
            spin1z_event = float(
                re.search('spin1z=(.+?)"', intrinsic_param_event).group(1)
            )
            spin2z_event = float(
                re.search('spin2z=(.+?)"', intrinsic_param_event).group(1)
            )
            event_params["spin1z"] = spin1z_event
            event_params["spin2z"] = spin2z_event
            (
                event_params["chi_eff"],
                event_params["chi_a"],
            ) = amrlib.transform_s1zs2z_chi_eff_chi_a(
                mass1_event, mass2_event, spin1z_event, spin2z_event
            )
        except AttributeError:
            event_params["spin1z"] = None
            event_params["spin2z"] = None
            logging.info("No Spin information found in event_info_dict")
            pass
        return event_params


def get_grid_info(rundir):
    results_dir = rundir + "/results/"
    all_xml = glob.glob(results_dir + "/ILE_iteration_*-MASS_SET_*-0.xml.gz")
    if len(all_xml) == 0:
        all_xml = glob.glob(
            results_dir + "/ILE_iteration_*-MASS_SET_*_0_.xml.gz"
        )
    print(f"Found {len(all_xml)} sample files")
    iterations = [
        xmlfile[
            xmlfile.find("ILE_iteration") : xmlfile.find("ILE_iteration")
            + len("ILE_iteration_0")
        ]
        for xmlfile in all_xml
    ]

    grid_levels = np.sort(np.unique(iterations))
    keys = {
        "mass1",
        "mass2",
        "spin1z",
        "spin2z",
        "chi_eff",
        "chi_a",
        "margll",
        "iteration_level",
        "filename",
    }
    data_dict = {key: [] for key in keys}
    check_spin = True
    for i, gl in enumerate(grid_levels):
        xml_files = glob.glob(results_dir + gl + "-MASS_SET_*-0.xml.gz")
        if len(xml_files) == 0:
            xml_files = glob.glob(results_dir + gl + "-MASS_SET_*_0_.xml.gz")
        print(f"Found {len(xml_files)} in grid_level {gl}")
        for xml_file in xml_files:
            xmldoc = utils.load_filename(
                xml_file, contenthandler=ligolw.LIGOLWContentHandler
            )
            new_tbl = lsctables.SnglInspiralTable.get_table(xmldoc)
            row = new_tbl[0]
            data_dict["filename"].append(xml_file)
            data_dict["mass1"].append(row.mass1)
            data_dict["mass2"].append(row.mass2)
            data_dict["margll"].append(row.snr)
            data_dict["iteration_level"].append(i)
            if check_spin:
                try:
                    data_dict["spin1z"].append(row.spin1z)
                    data_dict["spin2z"].append(row.spin2z)
                    (
                        chi_eff,
                        chi_a,
                    ) = amrlib.transform_s1zs2z_chi_eff_chi_a(
                        row.mass1, row.mass2, row.spin1z, row.spin2z
                    )
                    data_dict["chi_eff"].append(chi_eff)
                    data_dict["chi_a"].append(chi_a)
                except AttributeError:
                    check_spin = False
                    print("No spin information found in SnglInspiralTable")
    if data_dict["spin1z"] == []:
        del data_dict["spin1z"]
        del data_dict["spin2z"]
    data_dict = {key: np.array(data_dict[key]) for key in data_dict.keys()}
    return data_dict


def find_sigma(grid_data, param_list, sigma_factor, grid_level=None):
    """
    Find standard deviation of the gaussian at each grid point.
    Standand deviation at a given grid point is equal to half
    the separation between given grid point and its nearest
    neighbour grid point.
    """
    Sigma = {param: [] for param in param_list}
    for param in param_list:
        grid_param = np.array(grid_data[param])
        grid_iteration_level = grid_data["iteration_level"]
        grid_id = np.arange(len(grid_iteration_level))
        if grid_level is not None:
            grid_inds = grid_id[grid_iteration_level == grid_level]
            grid_param = np.array(grid_data[param])[grid_inds]
        for j in range(len(grid_param)):
            distance_array = np.array(
                [
                    abs(grid_param[j] - grid_param[i])
                    for i in range(len(grid_param))
                ]
            )
            distance_array = np.sort(distance_array[distance_array > 1e-5])
            distance = distance_array[0]
            Sigma[param] = np.append(
                Sigma[param], sigma_factor[param] * distance
            )

    return Sigma


def get_posterior_samples(
    grid_data,
    sigma,
    distance_coordinates_str,
    grid_level=None,
    spin_included=False,
    nsamples_per_grid=2000,
):
    """
    Generate posterior samples for params for the given grid_level
    """
    distance_coordinates = distance_coordinates_str.split("_")
    sample_dict = {}
    Margll_sel = grid_data["margll"]
    grid_it_level = grid_data["iteration_level"]
    grid_index_list = np.arange(len(grid_it_level))
    if grid_level is not None:
        grid_inds = grid_index_list[grid_it_level == grid_level]
        Margll_sel = grid_data["margll"][grid_inds]
        for param in distance_coordinates:
            grid_data[param] = grid_data[param][grid_inds]

    margL_normed = np.exp(Margll_sel - np.max(Margll_sel), dtype=np.float128)
    sum_margL_normed = np.sum(margL_normed)
    margL_normed /= sum_margL_normed
    seed = 12345
    random_state = np.random.RandomState(seed)
    N_mn = multinomial(
        nsamples_per_grid * len(margL_normed), margL_normed, seed=random_state
    )
    N = N_mn.rvs(1)[0]
    print(f"Number of samples {N}")
    grid_id = []
    all_random_samples = {param: [] for param in distance_coordinates}
    for i in range(len(margL_normed)):
        random_samples = {param: [] for param in distance_coordinates}
        for param in distance_coordinates:
            random_samples[param] = np.random.normal(
                loc=grid_data[param][i], scale=sigma[param][i], size=N[i]
            )
            all_random_samples[param] = np.append(
                all_random_samples[param], random_samples[param]
            )
        grid_id = np.append(np.full(N[i], i), grid_id)
    param1_samples = all_random_samples[distance_coordinates[0]]
    param2_samples = all_random_samples[distance_coordinates[1]]
    if spin_included:
        param3_samples = all_random_samples[distance_coordinates[2]]
        param4_samples = all_random_samples[distance_coordinates[3]]
    if distance_coordinates_str != "mu1_mu2_q_spin2z":
        mask = BOUND_CHECK_MASS[distance_coordinates_str](
            param1_samples, param2_samples
        )

        if spin_included:
            mask &= amrlib.check_spins(param3_samples)
            mask &= amrlib.check_spins(param4_samples)
        for param in distance_coordinates:
            all_random_samples[param] = all_random_samples[param][mask]

        param1_samples = all_random_samples[distance_coordinates[0]]
        param2_samples = all_random_samples[distance_coordinates[1]]
        if spin_included:
            param3_samples = all_random_samples[distance_coordinates[2]]
            param4_samples = all_random_samples[distance_coordinates[3]]

        m1_samples, m2_samples = INVERSE_TRANSFORMS_MASS[
            VALID_TRANSFORMS_MASS[
                frozenset(distance_coordinates_str.split("_"))
            ]
        ](param1_samples, param2_samples)
        prior = jacobians.PRIOR_MAP[distance_coordinates_str](
            param1_samples, param2_samples
        )
        sample_dict["mass1"] = m1_samples
        sample_dict["mass2"] = m2_samples

        sample_dict[distance_coordinates[0]] = param1_samples
        sample_dict[distance_coordinates[1]] = param2_samples
        sample_dict["prior"] = prior
        if spin_included:
            (
                spin1z_samples,
                spin2z_samples,
            ) = amrlib.transform_chi_eff_chi_a_s1zs2z(
                m1_samples, m2_samples, param3_samples, param4_samples
            )

            sample_dict["chi_eff"] = param3_samples
            sample_dict["chi_a"] = param4_samples
            sample_dict["spin1z"] = spin1z_samples
            sample_dict["spin2z"] = spin2z_samples
    else:
        mask = amrlib.check_q(param3_samples)
        mask &= amrlib.check_spins(param4_samples)
        mu1_samples = np.array(param3_samples[mask])
        mu2_samples = np.array(param4_samples[mask])
        q_samples = np.array(param1_samples[mask])
        spin2z_samples = np.array(param2_samples[mask])

        (
            m1_samples,
            m2_samples,
            spin1z_samples,
            spin2z_samples,
        ) = amrlib.transform_mu1mu2qs2z_m1m2s1zs2z(
            mu1_samples, mu2_samples, q_samples, spin2z_samples
        )

        chi_eff_samples, chi_a_samples = amrlib.transform_s1zs2z_chi_eff_chi_a(
            m1_samples,
            m2_samples,
            spin1z_samples,
            spin2z_samples,
        )
        mu1mu2qs2z_prior = jacobians.PRIOR_MAP[distance_coordinates_str](
            mu1_samples, mu2_samples, q_samples, spin2z_samples
        )
        sample_dict["mu1"] = mu1_samples
        sample_dict["mu2"] = mu2_samples
        sample_dict["q"] = q_samples
        sample_dict["spin2z"] = spin2z_samples
        sample_dict["spin1z"] = spin1z_samples
        sample_dict["chi_eff"] = chi_eff_samples
        sample_dict["chi_a"] = chi_a_samples
        sample_dict["mass1"] = m1_samples
        sample_dict["mass2"] = m2_samples
        sample_dict["prior"] = mu1mu2qs2z_prior
    sample_dict["grid_id"] = grid_id[mask]
    return sample_dict


def compute_evidence(
    grid_data,
    prior_boundary_dict,
    sigma,
    distance_coordinates_str,
    prior_function="uniform",
    grid_level=None,
):
    """
    Computes total evidence and category-wise evidence
    """
    prior_function = prior_function.lower()

    m_max_bank = prior_boundary_dict["m_max_bank"]
    m_min_bank = prior_boundary_dict["m_min_bank"]
    m_max_ns = prior_boundary_dict["m_max_ns"]

    count_evidence = count_bbh = count_bns = count_nsbh = 0

    distance_coordinates = list(sigma.keys())
    Margll_sel = grid_data["margll"]
    grid_it_level = grid_data["iteration_level"]
    grid_index_list = np.arange(len(grid_it_level))

    if grid_level is not None:
        grid_inds = grid_index_list[grid_it_level == grid_level]
        Margll_sel = grid_data["margll"][grid_inds]
        for param in distance_coordinates:
            grid_data[param] = grid_data[param][grid_inds]
    MargL = np.exp(Margll_sel, dtype=np.float128)
    seed = 12345
    np.random.seed(seed)

    for i in range(len(Margll_sel)):
        count_evidence_i = count_bbh_i = count_bns_i = count_nsbh_i = 0
        Nsamples = 10000
        random_samples = {param: [] for param in distance_coordinates}
        for param in distance_coordinates:
            random_samples[param] = np.random.normal(
                loc=grid_data[param][i], scale=sigma[param][i], size=Nsamples
            )
        param1_samples = random_samples[distance_coordinates[0]]
        param2_samples = random_samples[distance_coordinates[1]]
        if distance_coordinates_str != "mu1_mu2_q_spin2z":
            mask = BOUND_CHECK_MASS[distance_coordinates_str](
                param1_samples, param2_samples
            )
            for param in distance_coordinates:
                random_samples[param] = random_samples[param][mask]

            param1_samples = random_samples[distance_coordinates[0]]
            param2_samples = random_samples[distance_coordinates[1]]

            m1_samples, m2_samples = INVERSE_TRANSFORMS_MASS[
                VALID_TRANSFORMS_MASS[
                    frozenset(distance_coordinates_str.split("_"))
                ]
            ](param1_samples, param2_samples)
            uniform_prior = jacobians.PRIOR_MAP[distance_coordinates_str](
                param1_samples, param2_samples
            )
            if prior_function == "uniform":
                prior = uniform_prior
            elif prior_function == "salpeter":
                alpha = 2.35
                prior = m1_samples ** (-alpha) * uniform_prior
        else:
            param3_samples = random_samples[distance_coordinates[2]]
            param4_samples = random_samples[distance_coordinates[3]]

            mask = amrlib.check_q(param3_samples)
            mask &= amrlib.check_spins(param4_samples)
            mu1_samples = np.array(param3_samples[mask])
            mu2_samples = np.array(param4_samples[mask])
            q_samples = np.array(param1_samples[mask])
            spin2z_samples = np.array(param2_samples[mask])

            (
                m1_samples,
                m2_samples,
                spin1z_samples,
                spin2z_samples,
            ) = amrlib.transform_mu1mu2qs2z_m1m2s1zs2z(
                mu1_samples, mu2_samples, q_samples, spin2z_samples
            )

            prior = jacobians.PRIOR_MAP[distance_coordinates_str](
                mu1_samples, mu2_samples, q_samples, spin2z_samples
            )
        mchirp_samples, q_samples = amrlib.transform_m1m2_mcq(
            m1_samples, m2_samples
        )
        selected_indices_mask = q_samples <= prior_boundary_dict["q_max_bank"]
        selected_indices_mask &= (
            m1_samples <= prior_boundary_dict["m_max_bank"]
        )
        selected_indices_mask &= (
            m1_samples >= prior_boundary_dict["m_min_bank"]
        )
        selected_indices_mask &= (
            m2_samples <= prior_boundary_dict["m_max_bank"]
        )
        selected_indices_mask &= (
            m2_samples >= prior_boundary_dict["m_min_bank"]
        )
        if prior_function == "uniform":
            full_prior = prior[selected_indices_mask] / (
                (m_max_bank - m_min_bank)
                * (m1_samples[selected_indices_mask] - m_min_bank)
            )

        elif prior_function == "salpeter":
            full_prior = (
                prior[selected_indices_mask]
                * (1 - alpha)
                / (
                    (m_max_bank ** (1 - alpha) - m_min_bank ** (1 - alpha))
                    * (m1_samples[selected_indices_mask] - m_min_bank)
                )
            )

        Fofrandoms = MargL[i] * full_prior
        if not np.any(selected_indices_mask):
            continue
        try:
            random_F = np.random.uniform(
                0,
                np.amax(Fofrandoms),
                size=len(Fofrandoms),
            )
        except:
            continue
        Nsamples = len(Fofrandoms)
        area_sampled = (
            (
                np.amax(mchirp_samples[selected_indices_mask])
                - np.amin(mchirp_samples[selected_indices_mask])
            )
            * (
                np.amax(q_samples[selected_indices_mask])
                - np.amin(q_samples[selected_indices_mask])
            )
            * np.amax(Fofrandoms)
        )

        m1_samples_selected = m1_samples[selected_indices_mask]
        m2_samples_selected = m2_samples[selected_indices_mask]
        if prior_function == "uniform":
            prior_norm_bbh = 1.0 / (
                (m_max_bank - m_max_ns) * (m1_samples_selected - m_min_bank)
            )
            prior_norm_bns = 1.0 / (
                (m_max_ns - m_min_bank) * (m1_samples_selected - m_min_bank)
            )
            prior_norm_nsbh = 1.0 / (
                (m_max_bank - m_max_ns) * (m1_samples_selected - m_min_bank)
            )
            prior_norm_tot = 1.0 / (
                (m_max_bank - m_min_bank) * (m1_samples_selected - m_min_bank)
            )
        elif prior_function == "salpeter":
            prior_norm_bbh = (1.0 - alpha) / (
                (m_max_bank ** (1.0 - alpha) - m_max_ns ** (1.0 - alpha))
                * (m1_samples_selected - m_max_ns)
            )
            prior_norm_bns = (1.0 - alpha) / (
                (m_max_ns ** (1.0 - alpha) - m_min_bank ** (1.0 - alpha))
                * (m1_samples_selected - m_min_bank)
            )
            prior_norm_nsbh = np.asarray(
                (1.0 - alpha)
                / (
                    (m_max_bank ** (1.0 - alpha) - m_max_ns ** (1.0 - alpha))
                    * (m_max_ns - m_min_bank)
                )
            )
            prior_norm_nsbh = np.broadcast_to(
                prior_norm_nsbh, m1_samples_selected.shape
            )

            prior_norm_tot = (1.0 - alpha) / (
                (m_max_bank ** (1.0 - alpha) - m_min_bank ** (1.0 - alpha))
                * (m1_samples_selected - m_min_bank)
            )

        count_evidence_selected_samples = random_F <= Fofrandoms
        count_evidence_i = (
            np.count_nonzero(count_evidence_selected_samples)
            * area_sampled
            / Nsamples
        )

        m1_is_bh = m1_samples_selected > m_max_ns
        m1_is_ns = m1_samples_selected <= m_max_ns
        m2_is_ns = m2_samples_selected <= m_max_ns
        m2_is_bh = m2_samples_selected > m_max_ns

        count_bns_selected_samples = (
            count_evidence_selected_samples & m1_is_ns & m2_is_ns
        )
        count_bns_i = (
            np.sum(
                prior_norm_bns[count_bns_selected_samples]
                / prior_norm_tot[count_bns_selected_samples]
            )
            * area_sampled
            / Nsamples
        )

        count_nsbh_selected_samples = (
            count_evidence_selected_samples & m1_is_bh & m2_is_ns
        )
        count_nsbh_i = (
            np.sum(
                prior_norm_nsbh[count_nsbh_selected_samples]
                / prior_norm_tot[count_nsbh_selected_samples]
            )
            * area_sampled
            / Nsamples
        )

        count_bbh_selected_samples = (
            count_evidence_selected_samples & m1_is_bh & m2_is_bh
        )
        count_bbh_i = (
            np.sum(
                prior_norm_bbh[count_bbh_selected_samples]
                / prior_norm_tot[count_bbh_selected_samples]
            )
            * area_sampled
            / Nsamples
        )

        count_evidence += count_evidence_i
        count_bns += count_bns_i
        count_nsbh += count_nsbh_i
        count_bbh += count_bbh_i

    return {
        "evidence": count_evidence,
        "BNS": count_bns,
        "NSBH": count_nsbh,
        "BBH": count_bbh,
    }


def plot_grid(
    grid_data,
    param1,
    param2,
    plot_dir,
    event_info=None,
    grid_level=None,
    injection_info=None,
):
    """
    plot grid alignment for param1 and param2 and a specific grid level.

    Valid grid_level = 0,1,2,3,....None

    Valid param1 and param2 = mass1, mass2, mchirp, eta, spin1z, spin2z,
                              mu1, mu2, q, tau0, tau3, mtotal

    grid_level=None plots the grid point from all grid levels


    """
    logging.info(
        f"plotting grids for {param1} and {param2} on grid_level={grid_level}"
    )
    Margll = grid_data["margll"]
    grid_iteration_level = grid_data["iteration_level"]
    grid_id = np.arange(len(grid_iteration_level))
    if grid_level is not None:
        grid_inds = grid_id[grid_iteration_level == grid_level]
        data1 = grid_data[param1][grid_inds]
        data2 = grid_data[param2][grid_inds]
        weight = Margll[grid_inds]
    else:
        data1 = grid_data[param1]
        data2 = grid_data[param2]
        weight = Margll
    plt.figure()
    plt.scatter(
        data1,
        data2,
        c=weight,
        vmin=np.min(Margll),
        vmax=np.max(Margll),
    )
    plot_xmin = np.min(grid_data[param1])
    plot_xmax = np.max(grid_data[param1])
    plot_ymin = np.min(grid_data[param2])
    plot_ymax = np.max(grid_data[param2])
    if event_info is not None:
        plt.plot(
            event_info[param1],
            event_info[param2],
            "r*",
            label="pipeline_recovered",
        )
        plot_xmin = np.min([plot_xmin, event_info[param1]])
        plot_xmax = np.max([plot_xmax, event_info[param1]])
        plot_ymin = np.min([plot_ymin, event_info[param2]])
        plot_ymax = np.max([plot_ymax, event_info[param2]])
    if injection_info is not None:
        plt.plot(
            injection_info[param1],
            injection_info[param2],
            "m+",
            label="injected",
        )
        plot_xmin = np.min([plot_xmin, injection_info[param1]])
        plot_xmax = np.max([plot_xmax, injection_info[param1]])
        plot_ymin = np.min([plot_ymin, injection_info[param2]])
        plot_ymax = np.max([plot_ymax, injection_info[param2]])

    plt.xlabel(f"{param1}_d")
    plt.ylabel(f"{param2}_d")
    x_width = plot_xmax - plot_xmin
    y_width = plot_ymax - plot_ymin
    plt.xlim(
        plot_xmin - (0.1 * x_width),
        plot_xmax + (0.1 * x_width),
    )
    plt.ylim(
        plot_ymin - (0.1 * y_width),
        plot_ymax + (0.1 * y_width),
    )
    if grid_level is not None:
        plt.title("grid_level = " + str(grid_level))
    else:
        plt.title("all grids")
    plt.colorbar(label=r"$log(L_{marg})$")
    plt.legend()
    if grid_level is not None:
        filename = (
            f"{plot_dir}/grid_{param1}"
            f"_{param2}_iteration-{str(grid_level)}.png"
        )
    else:
        filename = f"{plot_dir}/grid_{param1}_{param2}_all.png"
    plt.savefig(filename)
    return


def plot_posterior(
    sample_dict,
    param,
    plot_dir,
    event_info=None,
    grid_level=None,
    injection_info=None,
):
    print(f"plotting posterior for {param} at grid_level={grid_level}")
    samples = sample_dict[param]

    if param in ["mass1", "mass2"]:
        prior = None
    else:
        prior = sample_dict["prior"]
        prior = prior / sum(prior)
    fig, ax = plt.subplots()

    lo, hi = np.percentile(samples, [0.1, 99.9])
    if event_info is not None:
        ax.axvline(
            x=event_info[param], color="red", label="pipeline_recovered"
        )
        lo = min(lo, event_info[param])
        hi = max(hi, event_info[param])
    if injection_info is not None:
        ax.axvline(x=injection_info[param], color="magenta", label="injected")
        lo = min(lo, injection_info[param])
        hi = max(hi, injection_info[param])
    lo = lo - (hi - lo) * 0.01
    hi = hi + (hi - lo) * 0.01
    bins = np.linspace(lo, hi, 50)
    ax.hist(
        samples,
        bins=bins,
        weights=prior,
        histtype="step",
        density=True,
        color="g",
    )
    ax.set_xlabel(f"{param}_d")
    ax.set_ylabel("posterior")
    ax.legend()
    ax.yaxis.set_ticks([])
    ax.set_xlim(lo, hi)
    if grid_level is not None:
        plt.title("grid_level = " + str(grid_level))
        filename = (
            f"{plot_dir}/posterior_detframe"
            f"{param}_iteration-{str(grid_level)}.png"
        )
    else:
        plt.title("all grids")
        filename = f"{plot_dir}/posterior_detframe_{param}_all.png"
    plt.savefig(filename)
    return


def plot_2d_posterior_with_grid(
    sample_dict,
    grid_data,
    distance_coordinates_str,
    plot_dir,
    grid_level=None,
    event_info=None,
    injection_info=None,
):
    distance_coordinates = distance_coordinates_str.split("_")
    param1_name = distance_coordinates[0]
    param2_name = distance_coordinates[1]
    grid_iteration_level = grid_data["iteration_level"]
    grid_id = np.arange(len(grid_iteration_level))
    if grid_level is not None:
        grid_inds = grid_id[grid_iteration_level == grid_level]
        data1 = grid_data[param1_name][grid_inds]
        data2 = grid_data[param2_name][grid_inds]
        weight = grid_data["margll"][grid_inds]
    else:
        data1 = grid_data[param1_name]
        data2 = grid_data[param2_name]
        weight = grid_data["margll"]
    all_weights = grid_data["margll"]
    plt.figure()
    plt.scatter(
        data1,
        data2,
        c=weight,
        vmin=np.min(all_weights),
        vmax=np.max(all_weights),
    )
    if event_info is not None:
        plt.plot(
            event_info[param1_name],
            event_info[param2_name],
            "r*",
            label="pipeline_recovered",
        )
    if injection_info is not None:
        plt.plot(
            injection_info[param1_name],
            injection_info[param2_name],
            "m+",
            label="injected",
        )

    plt.xlabel(f"{param1_name}_d")
    plt.ylabel(f"{param2_name}_d")
    plt.colorbar(label=r"$ln(L_{marg})$")
    plt.legend()

    samples1 = sample_dict[param1_name]
    samples2 = sample_dict[param2_name]
    prior = sample_dict["prior"]
    plt.hist2d(samples1, samples2, bins=50, weights=prior, density=True)
    if grid_level is not None:
        plt.title("grid_level = ", str(grid_level))
        filename = (
            f"{plot_dir}/{param1_name}_{param2_name}"
            f" _iteration-{str(grid_level)}.png"
        )
    else:
        plt.title("all grids")
        filename = f"{plot_dir}/{param1_name}_{param2_name}_all.png"
    plt.savefig(filename)
    return


def save_m1m2_posterior_samples(sample_dict, save_dir):
    print("saving poserior samples for intrinsic paramters in detector frame")
    filename = f"{save_dir}/intrinsic_posterior_samples_detframe.h5"
    f = h5py.File(filename, "w")
    f.create_dataset("mass1_d", data=sample_dict["mass1"])
    f.create_dataset("mass2_d", data=sample_dict["mass2"])
    f.create_dataset("grid_id", data=sample_dict["grid_id"])
    f.create_dataset("prior", data=sample_dict["prior"])

    if "spin1z" in sample_dict.keys():
        f.create_dataset("spin1z_d", data=sample_dict["spin1z"])
        f.create_dataset("spin2z_d", data=sample_dict["spin2z"])
    f.close()
    return
